from typing import Annotated
from fastapi import Request, Depends, Header


# HTMX Headers
HxRequest = Annotated[bool, Header(include_in_schema=False)]
# QuikUI Headers (this library)
QkVariant = Annotated[str | None, Header(include_in_schema=False)]


def request_if_html_response_needed(
    request: Request,
    # Header fields to detect with
    accept: Annotated[str | None, Header(include_in_schema=False)] = None,
    content_type: Annotated[str | None, Header(include_in_schema=False)] = None,
    hx_request: HxRequest = False,
    qk_variant: QkVariant = None,
) -> Request | None:
    """
    FastAPI dependency to detect if an ``HTMLResponse`` should be given to ``request``.
    Uses a serious of Header-based heuristics to determine if required.

    Returns:
        (Request | None): feedsback the request it was given if the detection triggers.
            Defaults to None (no HTML Request detected).
    """
    if hx_request or qk_variant:
        # We definitely know if any of these headers are present, it means respond w/ html
        return request

    elif content_type == "application/json":
        # Assume that if the request is JSON, the response should be too
        # NOTE: htmx never does this
        return None

    elif content_type == "application/jsonl":
        # Assume that if the request is JSONL,
        # the response should be streamed newline-delimited JSON
        # NOTE: htmx never does this
        return None

    elif accept == "text/event-stream":
        # Assume this is a browser-based SSE streaming call
        return request

    elif accept is not None and (
        accepted_types := list(t.strip() for t in accept.split(","))
    ):
        if any(t.startswith("text/html") for t in accepted_types):
            return request  # We have determined this is expecting HTML back

    # else: We haven't determined (according to above heuristics) that HTML is requested
    return None


RequestIfHtmlResponseNeeded = Annotated[
    Request | None, Depends(request_if_html_response_needed)
]



from collections.abc import AsyncGenerator, Callable
from typing import Any, Coroutine
from typing import ParamSpec, TypeVar

from fastapi import Response

P = ParamSpec("P")
T = TypeVar("T")
MaybeAsyncFunc = (
    Callable[P, T]
    | Callable[P, Coroutine[Any, Any, T] | Callable[P, AsyncGenerator[T]]]
)

# FastAPI-specific
FastApiHandler = Callable[P, Coroutine[None, None, T | Response]]
FastApiDecorator = Callable[[MaybeAsyncFunc[P, T]], FastApiHandler]



from fastapi import HTTPException, status
from typing import Any, Type


class BaseException(Exception):
    pass


class NoTemplateFound(BaseException, RuntimeError):
    def __init__(self, cls: Type, template_variant: str | None):
        if template_variant:
            template_name = f"{cls.__name__}.{template_variant}.html"
        else:
            template_name = f"{cls.__name__}.html"

        super().__init__(
            f"Component '{cls.__name__}' does not contain an environment with a template "
            f"named '{template_name}' in it, or subclass another component that can render itself."
        )


class HtmlResponseOnly(BaseException, HTTPException):
    def __init__(self):
        super().__init__(
            status_code=status.HTTP_406_NOT_ACCEPTABLE,
            detail="This route can only provide HTML responses. Please set Accept headers.",
        )


class ResponseNotRenderable(BaseException, ValueError):
    def __init__(self, result: Any):
        super().__init__(
            "Result type must either be an HTML-safe string, an instance of a BaseComponent "
            f"subclass, or an iterable of BaseComponent subclasses, not {type(result)}."
        )


__all__ = [
    HtmlResponseOnly.__name__,
    NoTemplateFound.__name__,
    ResponseNotRenderable.__name__,
]



from collections.abc import AsyncGenerator, AsyncIterator, Generator, Iterator


class EventStream(AsyncIterator[str]):
    def __init__(
        self,
        items: AsyncIterator[str] | Iterator[str],
        event: str | None = None,
        retry: int | None = None,
    ):
        if not isinstance(items, (AsyncGenerator, Generator)):
            raise ValueError(f"{type(items)} must be `aiter` or `iter`.")

        self.items = items
        self.event = event
        self.retry = retry

    def __aiter__(self) -> AsyncIterator[str]:
        return self

    def format_item(self, item: str) -> str:
        match self.event, self.retry:
            case None, None:
                return f"data: {item}\n\n"

            case str(event), None:
                return f"event: {event}\ndata: {item}\n\n"

            case None, int(retry):
                return f"data: {item}\nretry: {retry}\n\n"

            case str(event), int(retry):
                return f"event: {event}\ndata: {item}\nretry: {retry}\n\n"

    async def __anext__(self) -> str:
        if isinstance(self.items, AsyncGenerator):
            return self.format_item(await anext(self.items))

        return self.format_item(next(self.items))



import string
from collections.abc import Container
from enum import Enum
from functools import cache
from itertools import chain
from typing import (
    Annotated,
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterator,
    List,
    Literal,
    Self,
    Set,
)

from fastapi import Request
from fastapi.exceptions import RequestValidationError
from jinja2 import Environment, PackageLoader, Template
from jinja2 import TemplateNotFound as Jinja2TemplateNotFound
from markupsafe import Markup
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
    ValidationError,
    field_validator,
    model_serializer,
    model_validator,
)
from pydantic.fields import FieldInfo

from .exceptions import NoTemplateFound
from .utils import unflatten

# NOTE: https://html.spec.whatwg.org/multipage/syntax.html#attributes-2
VALID_ATTR_CHARS = set(string.printable) - set(""" "'`<>/=""")


def is_component(value: Any) -> bool:
    return isinstance(value, BaseComponent)


class CssClasses(RootModel):
    root: Set[str] = {}

    def update(self, other):
        if isinstance(other, CssClasses):
            self.root.update(other.root)
        else:
            self.root.update(other)

    def __bool__(self) -> bool:
        return bool(self.root)

    @model_validator(mode="before")
    @classmethod
    def convert_to_set(cls, val):
        if not val:
            return set()

        elif isinstance(val, str):
            return set(val.split(" "))

        elif isinstance(val, list):
            return set(val)

        return val

    @model_validator(mode="after")
    def validate_markup_safe(self):
        assert all(
            set(item).issubset(VALID_ATTR_CHARS) for item in self.root
        ), "Not in spec"
        return self

    @model_serializer()
    def serialize_css_classes(self) -> str:
        return " ".join(sorted(self.root))


class Attributes(RootModel):
    root: Dict[str, str | bool] = {}

    def __getitem__(self, key):
        return self.root.__getitem__(key)

    def get(self, key):
        return self.root.get(key)

    def __setitem__(self, key, val):
        return self.root.__setitem__(key, val)

    def update(self, other):
        if isinstance(other, Attributes):
            self.root.update(other.root)
        else:
            self.root.update(other)

    def __bool__(self) -> bool:
        return bool(self.root)

    @model_validator(mode="after")
    def validate_markup_safe(self):
        assert all(
            set(key).issubset(VALID_ATTR_CHARS) for key in self.root
        ), "Not in spec"
        assert all(
            set(value).issubset(set(string.printable)) for value in self.root.values()
        ), "Not in spec"
        return self

    @model_serializer()
    def serialize_html_attributes(self) -> str:
        return Markup(
            " ".join(
                (f'{k}="{v}"' if not isinstance(v, bool) else k)
                for k, v in self.root.items()
                # NOTE: Don't render non-"truthy" values e.g. `something=false`
                if v  # Relies on truthiness to render bools and strings correctly
            )
        )


class BaseComponent(BaseModel):
    """
    A subclass of `pydantic.BaseModel` that supports serialization to "safe" HTML. Serialization
    to HTML is either done by parsing a Jinja2 template (distributed with an associated package),
    or by explicitly overriding the `model_dump_html` function in a subclass. It is assumed that
    you take great care to properly handle user-generated content by "escaping" (making it safe).

    ```{warning}
    The way this library is designed does its best to maintain the property of always auto-escaping
    user-driven content. However, it is possible to configure your own components using this
    library to ignore that invariant. Either way, as a developer it is your job to ensure safety of
    the environment you are creating in your user's browser. For more information, check out:
    https://htmx.org/essays/web-security-basics-with-htmx/#always-use-an-auto-escaping-template-engine
    ```

    ```{note}
    Cannot use this model if you rely on special behavior with `ConfigDict(extras="allow")`, as this
    component overwrites the value of `__pydantic_extra__` with custom behavior for additonal attrs.
    ```
    """

    quikui_template_package_name: ClassVar[str] = "quikui"
    """The package that should be used to search for templates for components that do not
    override `model_dump_html`. If you want to define custom components using your own templates,
    override this in your base class and provide your package's name. Templates are stored in the
    `./{quikui_template_package_path}` directory as a package resource. Defaults to this package."""

    quikui_template_package_path: ClassVar[str] = "templates"
    """The directory inside this package that should be used to search for model templates to
    render. Defaults to `./templates`. Must be a package resource bundled with the package."""

    __quikui_component_name__: ClassVar[str | None] = None
    """To override the value of ``__quikui_component_name__`` when rendering the component."""

    __quikui_css_classes__: CssClasses
    """Add extra CSS classes to this component. Useful for integration with your design system."""

    __quikui_extra_attributes__: Attributes
    """Add extra attributes to this component. Exposed to template rendering."""

    # NOTE: Needed to fetch extra kwargs to models (will be discarded in `__init__`)
    __pydantic_extra__: dict[str, Any] = {}
    model_config = ConfigDict(extra="allow")

    @model_validator(mode="after")
    def parse_css_and_attrs(self):
        if hasattr(self, "__quikui_css_classes__") and hasattr(
            self, "__quikui_extra_attributes__"
        ):
            return self

        self.__quikui_css_classes__ = (
            CssClasses(root=self.__pydantic_extra__.pop("css", set()))
            if self.__pydantic_extra__
            else CssClasses()
        )

        if self.__pydantic_extra__:
            attrs = self.__pydantic_extra__.pop("attrs", dict())
            attrs.update(self.__pydantic_extra__)
            self.__quikui_extra_attributes__ = Attributes(root=attrs)

        else:
            self.__quikui_extra_attributes__ = Attributes()

        self.__pydantic_extra__ = {}  # Remove extras

        return self

    @classmethod
    @cache
    def quikui_environment(cls) -> Environment:
        """
        The environment to search for templates for this class and all it's subclasses.

        Returns:
            :class:`~jinja2.Environment`:
                The environment to search for template(s) to render this class with.

        ```{note}
        This method is cached since environments will typically not change during runtime.
        ```
        """
        env = Environment(
            loader=PackageLoader(
                package_name=cls.quikui_template_package_name,
                package_path=cls.quikui_template_package_path,
            ),
            autoescape=True,
        )
        # NOTE: Add our special filters here
        env.filters.update({"is_component": is_component})
        return env

    @classmethod
    def quikui_template(cls, template_variant: str | None = None) -> Template:
        """
        The template that should be used to render this model.

        Args:
            template_variant (str | None): Template type (file extension prepend) to find.
                This allows the use of custom templates for different scenarios, such as
                ``MyClass.list.html`` if ``template_type="list"``.
                Defaults to finding templates by their classname e.g. ``MyClass.html``.

        Returns:
            :class:`~jinja2.Template`: The template to render this model with.

        Raises:
            :class:`~quikui.NoTemplateFound`: If no template was found while recursing.

        ```{note}
        This method is not cached so updates to templates do not require reloading.
        ```

        ```{note}
        This classmethod is useful in combination with the ``template_variant`` keyword argument on
        the :func:`~quikui.render_component` decorator function in order to directly render a
        component.
        ```
        """
        template_class = cls
        while issubclass(template_class, BaseComponent):
            env = template_class.quikui_environment()

            try:
                return env.get_template(
                    f"{template_class.__name__}.{template_variant}.html"
                    if template_variant
                    else f"{template_class.__name__}.html"
                )

            except Jinja2TemplateNotFound:
                # If no template is found, recurse up the class heirarchy through `.__base__`
                # NOTE: Potentially dangerous if multi-class heirarchy exists, only uses first base
                assert template_class.__base__  # NOTE: For mypy
                template_class = template_class.__base__

        # NOTE: If we get to BaseComponent, there was some error in user's environment
        raise NoTemplateFound(cls, template_variant)

    def model_dump_html(
        self,
        include: Container | None = None,
        exclude: Container | None = None,
        template_variant: str | None = None,
        render_context: dict | None = None,
        **kwargs: dict,
    ) -> str:
        """
        Render this model to "safe" HTML. This default implementation assumes that a template
        exists for this model in `Class.quikui_environment` that can be used to serialize this
        model by it's top-level fields, including computed fields and any other fields that should
        be included.

        Args:

            include: Fields to include that would otherwise be skipped.
            exclude: Fields that should be skipped which would otherwise be included.
            template_variant (str | None): Template type (file extension) to find.
                This allows the use of custom templates for different scenarios, such as
                ``MyClass.list.html``. Defaults to finding normal ``MyClass.html`` templates.
            render_context (dict | None): Extra context to pass to ``.quikui_template(...).render``
            **kwargs: Any other attributes you want to pass directly to Jinja2 template rendering.

        Returns:
            (str): The rendered "safe" HTML that FastAPI will insert directly into a Response for
                this model.

        ```{note}
        You can override this if you can directly return "safe" html without using a template.
        ```

        ```{warning}
        This only gets the top-level current values, any recursion will be done via templating to
        ensure Markup safety context is bubbled up correctly. So, we do not pass down the further
        recursive context from `include` or `exclude` kwargs.
        ```
        """
        if not include:
            include = set()

        if not exclude:
            exclude = set()

        model_dict = dict(
            # NOTE: Ensure properties are in original form, not serialized
            (f, getattr(self, f))
            # NOTE: Ensure we get all public, computed, and any requested private fields...
            for f in chain(self.model_fields, self.model_computed_fields, include)
            #       ...but also allow skipping fields we don't need for template context.
            if f not in exclude
        )

        # NOTE: Because SQLModel doesn't run validators on DB load
        self.parse_css_and_attrs()

        if attrs := self.__quikui_extra_attributes__:
            model_dict["quikui_extra_attributes"] = attrs.model_dump()

        if css := self.__quikui_css_classes__:
            model_dict["quikui_css_classes"] = css.model_dump()

        model_dict["__quikui_component_name__"] = (
            self.__quikui_component_name__ or self.__class__.__name__
        )

        # NOTE: Feed the rest of the kwargs to this function directly to the template rendering
        model_dict.update(kwargs)

        return self.quikui_template(template_variant=template_variant).render(
            **model_dict,
            **(render_context or {}),
        )

    def __html__(self) -> str:
        """
        Serialize this model to "safe" HTML, using default settings for field include/exclude.
        This should not be overriden, except to modify how `model_dump_html` gets called by Jinja2.

        Returns:
            (str): The rendered "safe" HTML that FastAPI will insert directly into a Response for
                this model when using :func:`quikui.render_component`.

        ```{note}
        This allows `BaseComponent` models to automatically serialize themselves as HTML when used
        in a Jinga2 template. We can include private fields for the template engine this way.
        ```
        """
        return self.model_dump_html()


class Break(BaseComponent):
    def model_dump_html(self, **kwargs) -> str:
        return "<br>"


class Image(BaseComponent):
    source: str
    alt_text: str


class _SingleContentComponent(BaseComponent):
    content: str | BaseComponent

    @model_validator(mode="before")
    @classmethod
    def parse_content(cls, val) -> dict:
        if not isinstance(val, dict):
            return dict(content=val)

        return val

    def __init__(self, content: str | BaseComponent | None = None, **kwargs):
        # NOTE: Lets us do `cls(val)` instead of `cls(content=val)`
        if content:
            kwargs["content"] = content
        super().__init__(**kwargs)


class Heading(_SingleContentComponent):
    level: Annotated[int, Field(strict=True, ge=1, le=5, exclude=True)] = 1

    @property
    def __quikui_component_name__(self) -> str:
        return f"h{self.level}"


class Paragraph(_SingleContentComponent):
    __quikui_component_name__ = "p"


class Anchor(_SingleContentComponent):
    __quikui_component_name__ = "a"
    route: str


class Button(_SingleContentComponent):
    pass


class _MultiItemComponent(BaseComponent):
    items: list[str | BaseComponent] = []

    def __init__(self, *items, **kwargs):
        # NOTE: Let's us do `cls(*vals)` instead of `cls(items=vals)`
        if "items" not in kwargs:
            kwargs["items"] = list(items)

        elif len(items) > 0:
            raise ValueError(
                f"Cannot use `{self.__class__.__name__}(*items)` with `items=` kwarg."
            )

        super().__init__(**kwargs)


class Div(_MultiItemComponent):
    pass


class Span(_MultiItemComponent):
    pass


class ListItem(_SingleContentComponent):
    __quikui_component_name__ = "li"


class _ListComponent(_MultiItemComponent):
    """
    Component class that renders as a list component, with the inner content wrapped as `li`
    elements. Implemented in such as a way that you can get correct type hints on what it returns
    in the normal case where JSON is rendered, making it easy to use for both API and HTML modes.

    Usage example::

        >>> @app.get("/items")
        >>> @qk.render_component(wrapper=qk.UnorderedList)  # What it returns in HTML mode
        >>> def get_items() -> list[MyItem]:  # What it returns in JSON mode
        >>>     return session.exec(select(MyItem)).all()  # `MyItem` subclasses `BaseComponent`

    ```note
    You can add a common set of css classes or attributes to the inner list items by specifying
    ``item_css=set(...)`` and/or ``item_attributes=dict(...)`` to the class initialization. You
    can also manually construct the inner list item if you want to customize each item separately.
    ```
    """

    item_css: CssClasses | Callable[[int, ListItem], None] = Field(
        default_factory=CssClasses, exclude=True
    )
    item_attributes: Attributes | Callable[[int, ListItem], None] = Field(
        default_factory=Attributes, exclude=True
    )
    items: list[ListItem] = []

    @field_validator("items", mode="before")
    def add_li_if_missing(cls, items: list[str | BaseComponent]) -> list[ListItem]:
        return [
            ListItem(content=i) if not isinstance(i, ListItem) else i for i in items
        ]

    @model_validator(mode="after")
    def add_item_css_and_attributes(self):
        if isinstance(self.item_css, CssClasses):

            def apply_css_to_item(_: int, item: ListItem):
                item.__quikui_css_classes__.update(self.item_css)

        else:
            apply_css_to_item = self.item_css

        if isinstance(self.item_attributes, Attributes):

            def add_attributes_to_item(_: int, item: ListItem):
                item.__quikui_extra_attributes__.update(self.item_attributes)

        else:
            add_attributes_to_item = self.item_attributes

        for idx, item in enumerate(self.items):
            apply_css_to_item(idx, item)
            add_attributes_to_item(idx, item)

        return self


class UnorderedList(_ListComponent):
    __quikui_component_name__ = "ul"


class OrderedList(_ListComponent):
    __quikui_component_name__ = "ol"


class InputType(str, Enum):
    TEXT = "text"
    EMAIL = "email"
    PASSWORD = "password"
    SELECT = "select"
    SUBMIT = "submit"
    RESET = "reset"
    CHECKBOX = "checkbox"
    RADIO = "radio"

    def __str__(self) -> str:
        return self._value_


class Label(_SingleContentComponent):
    @model_validator(mode="before")
    @classmethod
    def convert_str_label(cls, val: str | dict) -> dict:
        if isinstance(val, str):
            return dict(content=val)

        return val


class FormInput(BaseComponent):
    type: Literal[InputType]
    name: str
    value: str | None = None
    label: Label | None = None
    add_break: bool = False
    required: bool = True

    @model_validator(mode="after")
    def add_id_to_label(self):
        if not (id := self.__quikui_extra_attributes__.get("id")):
            id = self.__quikui_extra_attributes__["id"] = self.name

        if self.label:
            self.label.__quikui_extra_attributes__["for"] = id

        return self

    @classmethod
    def from_model_field(cls, field_name: str, field_info: FieldInfo) -> "FormInput":
        form_attributes = field_info.json_schema_extra.get("form_attributes", {})

        return cls(name=field_name, **form_attributes)


class TextInput(FormInput):
    type: Literal[InputType] = InputType.TEXT


class EmailInput(FormInput):
    type: Literal[InputType] = InputType.EMAIL


class PasswordInput(FormInput):
    type: Literal[InputType] = InputType.PASSWORD


class InputOption(BaseModel):
    id: str
    value: str
    label: Label


class InputWithOptions(FormInput):
    options: list[InputOption] = []

    @classmethod
    def from_model_field(cls, field_name: str, field_info: FieldInfo) -> "FormInput":
        form_attributes = field_info.json_schema_extra.get("form_attributes", {})

        # NOTE: Override to allow adding options directly from enum field annotation
        if "options" not in form_attributes and issubclass(field_info.annotation, Enum):
            form_attributes["options"] = [
                InputOption(
                    id=option.name,
                    value=option.value,
                    label=Label(attrs={"for": option.name}, content=option.value),
                )
                for option in field_info.annotation
            ]

        return cls(name=field_name, **form_attributes)
        # NOTE: Can add options later via `self.options.extend(...)`


class SelectionInput(InputWithOptions):
    type: Literal[InputType] = InputType.SELECT
    # NOTE: Can set these after the initialization
    selected: str | None = None
    disabled: set[str] = {}

    @model_validator(mode="after")
    def check_selected_disabled(self) -> "SelectionInput":
        option_values = set(option.value for option in self.options)

        if self.selected and self.selected not in option_values:
            raise ValidationError("Selected must be one of the current options.")

        if not self.disabled.issubset(option_values):
            raise ValidationError("All disabled options should be current options.")

        return self


class CheckboxInput(FormInput):
    type: Literal[InputType] = InputType.CHECKBOX


class RadioInput(InputWithOptions):
    type: Literal[InputType] = InputType.RADIO


class SubmitForm(FormInput):
    type: Literal[InputType] = InputType.SUBMIT


class ResetForm(FormInput):
    type: Literal[InputType] = InputType.RESET


class Form(_MultiItemComponent):
    pass


class FormModel(BaseModel):
    """
    Model mix-in class that can produce a renderable form output.

    It is expected that you add the field metadata `form_type=Type[FormInput]`, and optionally
    you can include `field_attributes=dict(...)` to override the attributes of the generated
    FormInput class instance.

    ```note
    Does not subclass BaseComponent, you are meant to use `FormModel.create_form(...)` to produce
    `Form`, which is a `BaseComponent` object renderable to html
    ```
    """

    @classmethod
    def create_form_input(
        cls,
        field_name: str,
    ) -> FormInput:
        if not (field_info := cls.model_fields.get(field_name)):
            raise ValueError

        if not (input_class := field_info.json_schema_extra.get("form_type")):
            raise ValueError

        return input_class.from_model_field(field_name, field_info)

    @classmethod
    def create_form_items(cls, add_reset: bool = False) -> Iterator[BaseComponent]:
        """
        The sequence of form items that make up `Form.items` that is generated using
        `FormModel.create_form`

        You can override this method in your class to create a different layout pattern for your
        form. The default output yields:

            FormInput, Break, FormInput, Break, ..., Break, [ResetForm], SubmitForm
        """
        for field_name in cls.model_fields:
            yield cls.create_form_input(field_name)
            yield Break()

        if add_reset:
            yield ResetForm(name="reset")

        yield SubmitForm(name="submit")

    @classmethod
    def create_form(
        cls,
        add_reset: bool = False,
        css: List[str] = None,
        form_attrs: Dict[str, Any] = None,
        **attrs,
    ) -> Form:
        """
        Generate a `Form` component model which can be dumped to html
        """
        if form_attrs:
            attrs.update(form_attrs)

        return Form(
            css=css,
            attrs=attrs,
            items=list(cls.create_form_items(add_reset=add_reset)),
        )

    @classmethod
    async def as_form(cls, request: Request) -> Self:
        async with request.form() as form_data:
            model_data = unflatten(form_data)

        try:
            return cls.model_validate(model_data)

        except ValidationError as e:
            raise RequestValidationError(
                e.errors(include_input=True, include_url=True, include_context=True)
            )


__all__ = [
    BaseComponent.__name__,
    Heading.__name__,
    Paragraph.__name__,
    Break.__name__,
    Anchor.__name__,
    Button.__name__,
    Div.__name__,
    Span.__name__,
    ListItem.__name__,
    UnorderedList.__name__,
    OrderedList.__name__,
    Form.__name__,
    Label.__name__,
    FormInput.__name__,
    InputType.__name__,
    TextInput.__name__,
    EmailInput.__name__,
    PasswordInput.__name__,
    SelectionInput.__name__,
    InputOption.__name__,
    CheckboxInput.__name__,
    RadioInput.__name__,
    SubmitForm.__name__,
    FormModel.__name__,
]



import inspect
from functools import wraps
from typing import Callable, Iterable
from collections.abc import AsyncGenerator, Generator

from fastapi import Response
from fastapi.responses import HTMLResponse, StreamingResponse
from fastapi.templating import Jinja2Templates
from jinja2 import Environment, Template
from pydantic import BaseModel

from .components import BaseComponent, Div
from .dependencies import QkVariant, RequestIfHtmlResponseNeeded
from .exceptions import HtmlResponseOnly, ResponseNotRenderable
from .types import FastApiDecorator, FastApiHandler, MaybeAsyncFunc, P, T
from .utils import append_to_signature, execute_maybe_sync_func, get_response
from .sse import EventStream


def render_component(
    html_only: bool = False,
    template: Template | str | None = None,
    env: Environment | Jinja2Templates | None = None,
    wrapper: Callable[[Iterable[BaseComponent]], BaseComponent] | None = None,
    wrapper_kwargs: dict | None = None,
    streaming: bool = False,
) -> FastApiDecorator:
    """
    A decorator that should be used to automatically render an instance or sequence of
    :class:`~quikui.BaseComponent` subclass(es) into an :class:`~fastapi.responses.HTMLResponse`
    to respond to an incoming request that expects an HTML response (according to a heuristic).

    Args:
        html_only (bool):
            Whether this route should only accept Requests that expect an HTML response.
            Defaults to allowing both html and json responses, depending on heuristic detection.

        template (:class:`~jinja2.Template` | str | None):
            The template to use to render the model with using `result.model_dump()`.
            If it is a string, you **must** use the `env=` kwarg in tandem with it.
            Defaults to assuming return is a subclass of :class:`~quikui.BaseComponent`.

        env (:class:`~jinja2.Environment` | :class:`~fastapi.templating.Jinja2Templates` | None):
            The template environment that should be used to dynamically fetch a template to render
            with. Only is required if `template=` kwarg is a string value (the name of a template).

        wrapper:
            Function to use to wrap a sequence (e.g. ``list``) returned from a handler to
            transform it into an instance of :class:`~quikui.BaseComponent` for rendering.
            Defaults to :class:`~quikui.Div` for rendering a sequence result.

        **wrapper_kwargs:
            Keyword arguments to forward to constructing the model given in ``wrapper`` when
            rendering a sequence result (e.g. `return wrapper(*result, **wrapper_kwargs)`).

    Raises:
        :class:`~quikui.HtmlResponseOnly`:
            A 406 error if ``html_only`` is ``True`` and did not detect an HTML expected response.

        :class:`~quikui.ResponseNotRenderable`: If the result of evaluating the handler function is
            not an instance of str, :class:`~quikui.BaseComponent`, or
            a sequence of :class:`~quikui.BaseComponent`s, and no ``template`` is provided.

    Usage example::

        >>> import quikui as qk
        >>>
        >>> app = FastAPI()
        >>>
        >>> @app.get("/path")
        >>> @qk.render_component()
        >>> def get_something():
        >>>     return instance_of_basecomponent_subclass  # Automatically converted to HTMLResponse

    ```{warning}
    If using the ``template`` keyword argument, please note that the result model is dumped using
    the normal Pydantic `.model_dump` method and therefore will not contain any extra attributes
    or classes that QuikUI provides.
    ```

    ```{warning}
    Errors are only raised when the decorated function is called.
    Any unexpected error will generate a 500 Server Error in your FastAPI app.
    ```
    """

    # NOTE: We must create a getter for the template so that it is dynamically fetched when needed
    if isinstance(template, str):
        if env is None:
            raise AssertionError("`env=` must be set if `template=` is a str.")

        def get_template():
            return env.get_template(template)

    else:

        def get_template():
            return template  # type is `Template | None`

    def decorator(func: MaybeAsyncFunc[P, T]) -> FastApiHandler:
        @wraps(func)
        async def wrapper_render_if_html_requested(
            *args: P.args,
            __html_request: RequestIfHtmlResponseNeeded,
            qk_variant: QkVariant = None,
            **kwargs: P.kwargs,
        ) -> T | Response:
            if html_only and __html_request is None:
                raise HtmlResponseOnly()

            # NOTE: Short-circut to return response directly if user returns one
            if isinstance(
                result := await execute_maybe_sync_func(func, *args, **kwargs),
                Response,
            ):
                return result

            elif __html_request is None:
                # NOTE: Fallback on FastAPI-native processing of models if not rendering HTML

                if not streaming:
                    return result

                elif isinstance(result, AsyncGenerator):
                    model_iter = (
                        item.model_dump_json() + "\n" async for item in result
                    )

                elif isinstance(result, Generator):
                    model_iter = (item.model_dump_json() + "\n" for item in result)

                else:
                    raise RuntimeError(f"{type(result)} is not iter/aiter.")

                return StreamingResponse(
                    model_iter,
                    # NOTE: Content is encoded as newline-delimited JSON
                    headers={"Content-Type": "application/jsonl"},
                )

            # NOTE: Dependency resolves to a `Request` if we've made it this far
            request = __html_request

            # Render model as HTML using Template
            if (response_template := get_template()) and isinstance(
                result, (BaseModel, dict)
            ):
                result = response_template.render(
                    **(
                        result.model_dump() if isinstance(result, BaseModel) else result
                    ),
                    request=request,
                    url_for=request.url_for,
                )

            # Render sequence of models as HTML using Template, wrapped in another element
            elif (
                response_template
                and isinstance(result, (tuple, list))
                and all(isinstance(r, (BaseModel, dict)) for r in result)
            ):
                result = (wrapper if wrapper else Div)(  # type: ignore[operator]
                    *(
                        response_template.render(
                            **(r.model_dump() if isinstance(r, BaseModel) else r),
                            request=request,
                            url_for=request.url_for,
                        )
                        for r in result
                    ),
                    **(wrapper_kwargs or {}),
                )

            # Wrap sequence of models into another element to process further
            elif isinstance(result, (tuple, list)) and all(
                isinstance(r, (BaseComponent, str)) for r in result
            ):
                result = (wrapper if wrapper else Div)(  # type: ignore[operator]
                    *result,
                    **(wrapper_kwargs or {}),
                )

            # Assume sequence of models is a generator,
            # and stream response (rendering each model as HTML)
            elif streaming:
                if isinstance(result, AsyncGenerator):
                    model_iter = (
                        item.model_dump_html(  # type: ignore[assignment]
                            template_variant=qk_variant,
                            render_context=dict(
                                request=request, url_for=request.url_for
                            ),
                        )
                        async for item in result
                    )

                elif isinstance(result, Generator):
                    model_iter = (
                        item.model_dump_html(  # type: ignore[assignment]
                            template_variant=qk_variant,
                            render_context=dict(
                                request=request, url_for=request.url_for
                            ),
                        )
                        for item in result
                    )

                else:
                    raise RuntimeError(f"{type(result)} is not iter/aiter.")

                return StreamingResponse(
                    EventStream(model_iter),
                    headers={"Content-Type": "text/event-stream"},
                )

            # Raise if we can't handle the rest of this
            elif not isinstance(result, (BaseComponent, str)):
                # NOTE: Should not happen if library is used properly
                raise ResponseNotRenderable(result)

            # We have a model to render to HTML (or converted to one above)
            if isinstance(result, BaseComponent):
                result = result.model_dump_html(  # type: ignore[assignment]
                    template_variant=qk_variant,
                    render_context=dict(request=request, url_for=request.url_for),
                )

            # else: `result` is assumed to be HTML str now (NOTE: could be unsafe)

            if response := get_response(kwargs):
                return HTMLResponse(result, headers=response.headers)

            else:
                return HTMLResponse(result)

        # NOTE: Ensure our html response detection dependency is included
        return append_to_signature(
            wrapper_render_if_html_requested,
            inspect.Parameter(
                "__html_request",
                inspect.Parameter.KEYWORD_ONLY,
                annotation=RequestIfHtmlResponseNeeded,
            ),
            inspect.Parameter(
                "qk_variant",
                inspect.Parameter.KEYWORD_ONLY,
                annotation=QkVariant,
                default=None,
            ),
        )

    return decorator



from .components import *
from .components import __all__ as __all_components__
from .decorators import render_component
from .dependencies import QkVariant, RequestIfHtmlResponseNeeded
from .exceptions import __all__ as __all_exceptions__

__all__ = [
    "QkVariant",  # NOTE: types don't have `.__name__`
    "RequestIfHtmlResponseNeeded",  # NOTE: types don't have `.__name__`
    render_component.__name__,
    *__all_components__,
    *__all_exceptions__,
]



import asyncio
import random
from collections.abc import AsyncIterator
from datetime import datetime, timezone, tzinfo
from enum import Enum

import quikui as qk
from fastapi import Depends, FastAPI, Form, Request
from fastapi.templating import Jinja2Templates
from pydantic import BaseModel, Field

app = FastAPI()

# You can use your existing the templates the same
templates = Jinja2Templates(directory="example/templates")


# NOTE: It is recommended to exclude `html_only=True` routes from the schema
@app.get("/", include_in_schema=False)
@app.get("/index.html", include_in_schema=False)
# In fact, `template=` is a nice way to render pages directly without needing a custom Component
@qk.render_component(html_only=True, template="CustomPage.html", env=templates)
async def index():
    # NOTE: When using `template=`, you can just directly return a dict or BaseModel
    return dict(
        title="Basic Demo App",
        content=[
            qk.Div(
                qk.Heading("Basic Component Demo"),
                qk.Paragraph("This is a paragraph element."),
                qk.Div(
                    qk.Paragraph("This is a paragraph inside a div."),
                    qk.Paragraph("This is another paragraph inside a div."),
                    qk.Paragraph(
                        content=qk.Span(
                            "This is a span ",
                            qk.Anchor(
                                "with a link",
                                route="https://google.com",
                                # You can add extra html attributes via kwargs
                                target="_blank",  # NOTE: Open link in a new tab
                            ),
                            " to something inside that same div.",
                        )
                    ),
                    # Can add custom CSS classes to any component via `css=...`
                    css="my-3",
                ),
                qk.Paragraph(
                    qk.Span(
                        "This is another span with a link",
                        qk.Anchor(
                            " to another page",
                            route="/another-page",
                        ),
                        ", outside the div.",
                    ),
                ),
            )
        ],
    )


@app.get("/another-page")
# NOTE: You can supply a template directly, but be aware it will not automatically update
#       (Try modifying `CustomPage.html` and refreshing both this page and the previous)
@qk.render_component(html_only=True, template=templates.get_template("CustomPage.html"))
def another_page():
    return dict(
        title="A page with dynamic content",
        content=[
            qk.Heading("Using HTMX"),
            qk.Paragraph(
                "This button uses HTMX to dynamically fetch content"
                " from the server using a GET request."
            ),
            qk.Button(
                "Get Dynamic Content",
                # can also add extra attributes via `attrs=dict(...)` kwarg
                # NOTE: This is useful for when attrs have `-` in them, or are protected keywords
                attrs={"hx-get": "/dynamic", "type": "button", "hx-swap": "outerHTML"},
            ),
            qk.Button(
                "Get Form",
                attrs={"hx-get": "/form", "type": "button", "hx-swap": "outerHTML"},
            ),
        ],
    )


# NOTE: The components used do not have to be so fine-grained,
class Component(qk.BaseComponent):
    """To make a renderable component, just subclass BaseComponent"""

    # NOTE: If you override `quikui_template_package_name` class variable with your own
    #       package or app, and that contains a `/templates` folder that has a template
    #       with the name of this class and the `.html` extension, then it will "auto-render"
    quikui_template_package_name = "example"


class CustomComponent(Component):
    # NOTE: It is recommended to create a base class that your other components inherit from

    # You can add whatever fields you'd like to your models
    text: str = "Some random gibberish!"


@app.get("/dynamic")
@qk.render_component()  # NOTE: Allows both w/ html and json response modes when `html_only=False`
def dynamic_content():
    return [
        CustomComponent(text=random.choice(["Select", "Dynamic", "Content"]))
        for _ in range(random.randint(1, 10))
    ]


class CarTypes(Enum):
    NONE_SELECTED = "Select a Car"
    FIAT = "Fiat"
    DODGE = "Dodge"


# NOTE: This is not a component!
class CustomForm(qk.FormModel):
    username: str = Field(
        form_type=qk.TextInput,
        form_attributes=dict(label="Username:", add_break=True),
    )
    email: str = Field(
        form_type=qk.EmailInput,
        form_attributes=dict(label="Email:", add_break=True),
    )
    password: str = Field(
        form_type=qk.PasswordInput,
        form_attributes=dict(label="Password:", add_break=True),
    )
    save_password: bool = Field(
        form_type=qk.CheckboxInput,
        form_attributes=dict(label="Save Password?"),
    )
    car_choice: CarTypes = Field(
        form_type=qk.RadioInput,
        form_attributes=dict(label="First Choice:"),
    )
    backup_choice: CarTypes = Field(
        form_type=qk.SelectionInput,
        form_attributes=dict(
            label="Second Choice:",
            selected=CarTypes.NONE_SELECTED,
            disabled={CarTypes.NONE_SELECTED},
        ),
    )

    # NOTE: Override `.create_form_items()` with custom form item generator if desired


@app.get("/form")
@qk.render_component(html_only=True)  # NOTE: Only need this page to fetch the form
def form_page():
    return CustomForm.create_form(
        id="a-form", form_attrs={"hx-post": "/completed-form"}
    )


@app.post("/completed-form")
@qk.render_component(
    # You can give a "wrapper" component to use for wrapping an iterable result in html render mode
    wrapper=qk.UnorderedList,  # NOTE: By default uses `qk.Div`
    # NOTE: Can add additional keyword arguments to initializing `wrapper`
    wrapper_kwargs=dict(
        item_css={"my-5"},
        item_attributes=dict(something="else"),
    ),
)
async def receive_form(form: CustomForm = Depends(CustomForm.as_form)):
    # NOTE: The `form_handler` dependency will handle parsing and unflattening native HTML Forms
    return [f"{field}: {value}" for field, value in form.model_dump().items()]


class Datapoint(Component):
    time: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))
    data: float = Field(default_factory=random.random)


data: list[Datapoint] = list()


@app.get("/dashboard")
@qk.render_component(html_only=True, template=templates.get_template("Dashboard.html"))
async def dashboard():
    return dict(data=data)


@app.get("/stream-data")
# NOTE: Use `streaming=True` to encode the result as a StreamingResponse (using SSE EventSource for HTML)
@qk.render_component(streaming=True)
# NOTE: QuikUI handles the conversion to SSE-compatible streaming e.g. `data:{{ html }}\n\n`
async def stream_live_data() -> "Datapoint":
    while True:
        await asyncio.sleep(0.25)
        pt = Datapoint()
        data.append(pt)
        yield pt


class Notification(Component):
    msg: str


@app.get("/notifications")
@qk.render_component(streaming=True)
# NOTE: streaming responses don't have to be endless iterators
async def notifications() -> "Notification":
    yield Notification(msg="Wait 5 seconds for a notification")
    await asyncio.sleep(5.0)
    yield Notification(msg="Wait another 5 seconds for another")
    await asyncio.sleep(5.0)
    yield Notification(msg="No more notifications!")
    await asyncio.sleep(50.0)
    yield Notification(msg="Just kidding, htmx-sse auto-reconnects...")
